import MetaTrader5 as mt5

mt5tf_map = {
    "1M": mt5.TIMEFRAME_M1,
    "2M": mt5.TIMEFRAME_M2,
    "3M": mt5.TIMEFRAME_M3,
    "4M": mt5.TIMEFRAME_M4,
    "5M": mt5.TIMEFRAME_M5,
    "6M": mt5.TIMEFRAME_M6,
    "10M": mt5.TIMEFRAME_M10,
    "12M": mt5.TIMEFRAME_M12,
    "15M": mt5.TIMEFRAME_M15,
    "30M": mt5.TIMEFRAME_M30,
    "1H": mt5.TIMEFRAME_H1,
    "2H": mt5.TIMEFRAME_H2,
    "3H": mt5.TIMEFRAME_H3,
    "4H": mt5.TIMEFRAME_H4,
    "6H": mt5.TIMEFRAME_H6,
    "8H": mt5.TIMEFRAME_H8,
    "12H": mt5.TIMEFRAME_H12,
    "1D": mt5.TIMEFRAME_D1,
    "1W": mt5.TIMEFRAME_W1,
    "1MN": mt5.TIMEFRAME_MN1,
}

order_types_map = {
    # Immediate Execution w/ specified parameters (market order)
    "MARKET": mt5.TRADE_ACTION_DEAL,
    # Order with conditions (pending order)
    "LIMIT": mt5.TRADE_ACTION_PENDING,
    # Modify Stoploss & Take Profit values for opened order
    "ADD_SLTP": mt5.TRADE_ACTION_SLTP,
    # Modify parameters of previously placed order
    "MODIFY": mt5.TRADE_ACTION_MODIFY,
    # delete pending order
    "DEL_PENDING": mt5.TRADE_ACTION_REMOVE,
    # square off an opened position, (close opened order by opposite one)
    "CLOSE": mt5.TRADE_ACTION_CLOSE_BY,
}

position_type = {
    "MARKET_BUY": mt5.ORDER_TYPE_BUY,
    "MARKET_SELL": mt5.ORDER_TYPE_SELL,
    "LIMIT_BUY": mt5.ORDER_TYPE_BUY_LIMIT,
    "LIMIT_SELL": mt5.ORDER_TYPE_SELL_LIMIT,
    "LIMIT_BUY_STOP": mt5.ORDER_TYPE_BUY_STOP,
    "LIMIT_SELL_STOP": mt5.ORDER_TYPE_SELL_STOP,
    # "LIMIT_BUY_STOP_LIMIT": mt5.ORDER_TYPE_BUY_STOP_LIMIT,
    # "LIMIT_SELL_STOP_LIMIT": mt5.ORDER_TYPE_SELL_STOP_LIMIT,
    # "SQ_OFF": mt5.ORDER_TYPE_CLOSE_BY,
}

filling_type = {
    "FILL_FULL": mt5.ORDER_FILLING_FOK,
    "FILL_AVAILABLE": mt5.ORDER_FILLING_IOC,
    # "FILL_RETURN": mt5.ORDER_FILLING_RETURN
}

time_type = {
    # order stays in the queue until it is manually canceled
    "VALID_TILL_CANCELED": mt5.ORDER_TIME_GTC,
    # order is active only during the current trading day
    "VALID_DAY": mt5.ORDER_TIME_DAY,
    # order is active until the specified date
    "VALID_TIME_SPECIFIED": mt5.ORDER_TIME_SPECIFIED,
    # order is active until 23:59:59 of the specified day. If this time appears to be out of a trading session, the expiration is processed at the nearest trading time.
    "VALID_TIME_SPECIFIED_DAY": mt5.ORDER_TIME_SPECIFIED_DAY,
}


from datetime import datetime
import pandas as pd

from metrapy.defaults import defaults

def get_candles_df(candles, timezone=defaults.timezone):
    _candles = pd.DataFrame(
        candles,
        # columns=['Datetime', 'Open', 'High', 'Low', 'Close', 'tick_volume', 'spread', 'real_volume']
    )

    _candles["Datetime"] = _candles["time"].map(
        lambda x: datetime.fromtimestamp(x, timezone)
    )

    _candles = _candles.set_index("Datetime")

    return _candles


def filter_props(obj, keys):
    t = {}
    for key in keys:
        t[key] = obj[key]
    return t



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="metrapy",  # This is the name of the package
    version="0.1.3",  # The initial release version
    author="pavittarx",  # Full name of the author
    description="batteries included wrapper utility for MetaTrader5's python integration",
    long_description=long_description,  # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],  # Information to filter the project on PyPi website
    python_requires=">=3.6",  # Minimum version requirement of the package
    py_modules=["connector", "sync"],  # Name of the python package
    # package_dir={"": "metrapy"},  # Directory of the source code of the package
    install_requires=["MetaTrader5"],  # Install other dependencies if any
)



from threading import Timer

from datetime import datetime
import pytz

from metrapy.defaults import defaults

def get_current_time(timezone=defaults.timezone):
    return datetime.now().astimezone(timezone)

def get_sync_secs(timeframe):
    now = get_current_time()

    diff_secs = 0

    if type(timeframe) == int:
        return timeframe

    if timeframe == "1s":
        diff_secs = 1

    if timeframe == "1M":
        diff_secs = 60 - now.second

    if timeframe == "2M":
        diff_secs = ((2 - (now.minute % 2)) * 60) - now.second
        # print("diff ", timeframe, now.minute, now.minute%2 , (now.minute % 2) * 60)
        # print("left over: ", now.second, now.minute%2 * 60 + 60 - now.second)

    if timeframe == "3M":
        diff_secs = ((3 - (now.minute % 3)) * 60) - now.second

    if timeframe == "4M":
        diff_secs = ((4 - (now.minute % 4)) * 60) - now.second

    if timeframe == "5M":
        diff_secs = ((5 - (now.minute % 5)) * 60) - now.second

    if timeframe == "6M":
        diff_secs = ((6 - (now.minute % 6)) * 60) - now.second

    if timeframe == "15M":
        diff_secs = ((15 - (now.minute % 15)) * 60) - now.second

    if timeframe == "30M":
        diff_secs = ((30 - (now.minute % 30)) * 60) - now.second

    if timeframe == "1H":
        diff_secs = 3600 - (now.minute * 60) - now.second

    if timeframe == "4H":
        diff_secs = (4 - (now.hour % 4)) * 3600 - (now.minute * 60) - now.second

    if timeframe == "1D":
        diff_secs = 86400 - (now.hour * 3600) - (now.minute * 60) - now.second

    return diff_secs

def poller(func, timeframe, adjust_secs=0):
    def wrapper():
        poller(func, timeframe, adjust_secs)
        func()

    secs = get_sync_secs(timeframe) + adjust_secs
    t = Timer(secs, wrapper)
    t.start()
    return t



from metrapy.sync import get_current_time
from metrapy.maps import (
    mt5tf_map,
    order_types_map,
    position_type,
    filling_type,
    time_type,
)

from metrapy.defaults import defaults
from metrapy.utils import get_candles_df, filter_props

import MetaTrader5 as mt5
import pandas as pd
import uuid

import logging

class MT5:
    _mt5 = mt5
    magic = None
    timezone = defaults.timezone
    connected = False
    debug = False
    maps = {
        'timeframe': mt5tf_map,
        'order': order_types_map,
        'position': position_type,
        'filling': filling_type,
        'time': time_type
    }

    def __init__(self, params, opts={}):
        if "login" not in params:
            raise Exception("Please provide MT5 login.")

        if "pass" not in params:
            raise Exception("Please provide MT5 password.")

        if "server" not in params:
            raise Exception("MT5 Server Name is missing. Please provide.")

        if "path" not in params:
            print(
                "MT5 binary (terminal) installation path is mising. \n Using Default: C:/Program Files/MetaTrader 5/terminal64.exe"
            )

        if type(params["login"]) != int:
            raise Exception("login parameter must be a number")

        self.connected = mt5.initialize(
            path=params["path"] if "path" in params else defaults.mt5_path,
            server=params["server"],
            login=params["login"],
            password=params["pass"],
        )

        if not self.connected:
            print("Error while trying to connect to MT5")
            print(params)
            quit()

        if "timezone" in opts:
            self.timezone = opts["timezone"]

        if "debug" in opts:
            self.debug = True

        if "magic" in params:
            self.magic = params['magic_id']

    def get_account_info(self):
        return mt5.account_info()._asdict()

    def get_tickers(self, params={}):
        if "ticker" in params:
            sym = mt5.symbol_info(params["ticker"])
            return filter_props(sym._asdict(), params["filter_props"])

        if "group" in params:
            symbols = mt5.symbols_get(group=params["group"])
        else:
            symbols = mt5.symbols_get()
        
        tickers = []

        for sym in symbols:
            if "filter_props" in params:
                tickers.append(filter_props(sym._asdict(), params["filter_props"]))
                continue

            tickers.append(sym._asdict())

        return tickers

    def get_tickers_count(self, params={}):
        if "group" in params:
            return mt5.symbols_get(group=params["group"])

        return mt5.symbols_total()

    def get_candles(self, params):
        if "ticker" not in params:
            raise Exception("Please provide ticker information")

        if "timeframe" not in params:
            raise Exception(
                "Please provide timeframe for which candles are to be fetched"
            )

        if "count" not in params:
            raise Exception("Please provide count of candles to be fetched.")

        candles = mt5.copy_rates_from_pos(
            params["ticker"],
            mt5tf_map[params["timeframe"]],
            params["start"] if "start" in params else 0,
            params["count"],
        )

        if candles is None:
            return {
                "candles": None,
                "last_error": mt5.last_error(),
                "params": params,
            }

        if self.debug:
            print(
                "Candles Fetched: ",
                params["ticker"],
                get_current_time().strftime("%m/%d/%Y, %H:%M:%S", len(candles)),
            )

        candles_df = get_candles_df(candles, timezone= params['timezone'] if "timezone" in params else None)

        return {"candles": candles, "last_error": None, "params": params, 'candles_df': candles_df}

    def get_last_candle(self, params):
        if "ticker" not in params:
                raise Exception("Please provide ticker information")
        
        if "timeframe" not in params:
                raise Exception(
                    "Please provide timeframe for which candles are to be fetched"
                )
        
        candle_raw = mt5.copy_rates_from_pos(
                params["ticker"],
                mt5tf_map[params["timeframe"]],
                0,
                1,
            )
        
        if candle_raw is None:
                return {
                    "candle": None,
                    "last_error": mt5.last_error(),
                    "params": params,
                }
        
        if self.debug:
                print(
                    "Candles Fetched: ",
                    params["ticker"],
                    get_current_time().strftime("%m/%d/%Y, %H:%M:%S", len(candle_raw)),
                )
        
        candle = get_candles_df(candle_raw, timezone= params['timezone'] if "timezone" in params else None)
        
        return {"candle_raw": candle[0] if len(candle) > 0 else None, "last_error": None, "params": params, 'candles_df': candle.iloc[0] if candle else None}

    def get_orders(self, params):
        request = {}

        if "ticker" in params:
            request["symbol"] = params["ticker"]

        if "ticket" in params:
            request["ticket"] = params["ticket"]

        if "group" in params:
            request["group"] = params["group"]

        orders = mt5.orders_get(**request)

        if orders is None:
            return {
                'orders': None,
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }
        
        orders_list = []

        for o in orders:
            orders_list.append(o._asdict())

        return {
            'orders': orders_list,
            'count': len(orders_list),
            'last_error': None,
            'params': params,
            '_request': request
        }

    def get_positions(self, params):
        request = {}

        if "ticker" in params:
            request["symbol"] = params["ticker"]

        if "ticket" in params:
            request["ticket"] = params["ticket"]

        if "group" in params:
            request["group"] = params["group"]

        positions = mt5.positions_get(**request)

        if positions is None:
            return {
                'orders': None,
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }
        
        pos_list = []

        for pos in positions:
            pos_list.append(pos._asdict())

        return {
            'positions': pos_list,
            'count': len(pos_list),
            'last_error': None,
            'params': params,
            '_request': request
        }

    def get_history(self, params):

        request = {}
        
        if 'ticket' in params:
            ticket = params["ticket"]

            if type(ticket) != int:
                raise Exception("ticket parameter should be an integer.")

            request['ticket'] = ticket

        if "position" in params:

            pos = params["position"]

            if type(pos) != int:
                raise Exception("position parameter should be an integer.")

            request['position'] = pos

        if "start" in params:
            request["date_from"] = params["start"]

        if "end" in  params:
            request["date_to"] = params["end"]

        if "group" in params:
            request["group"] = params["group"]

        orders = mt5.history_orders_get(**request)

        if orders is None:
            return {
                'orders': None,
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }
        
        orders_list = []

        for o in orders:
            orders_list.append(o._asdict())

        return {
            'orders': orders_list,
            'count': len(orders_list),
            'last_error': None,
            'params': params,
            '_request': request
        }
    
    def get_history_deals(self, params):

        request = {}
        
        if 'ticket' in params:
            ticket = params["ticket"]

            if type(ticket) != int:
                raise Exception("ticket parameter should be an integer.")

            request['ticket'] = ticket

        if "position" in params:

            pos = params["position"]

            if type(pos) != int:
                raise Exception("position parameter should be an integer.")

            request['position'] = pos

        if "start" in params:
            request["date_from"] = params["start"]

        if "end" in  params:
            request["date_to"] = params["end"]

        if "group" in params:
            request["group"] = params["group"]

        deals = mt5.history_deals_get(**request)

        if deals is None:
            return {
                'deals': None,
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }
        
        deals_list = []

        for deal in deals:
            deals_list.append(deal._asdict())

        return {
            'orders': deals_list,
            'count': len(deals_list),
            'last_error': None,
            'params': params,
            '_request': request
        }

    def place_order(self, params):
        id = uuid.uuid4().int & (1 << 64) - 1

        def get_order_type():
            type = params["order_type"]

            if params["trade_type"] == "short":
                type = type + "_SELL"

            if params["trade_type"] == "long":
                type = type + "_BUY"

            if "stop_order" in params and params["stop_order"]:
                type = type + "_STOP"

            if "stop_limit_order" in params and params["stop_limit_order"]:
                type = type + "_STOP_LIMIT"

            return type

        request = {
            "magic": self.magic,
            "symbol": params["ticker"],
            "action": order_types_map[params["order_type"]],
        }

        if "trade_type" in params:
            request["type"] = position_type[get_order_type()]

        if "filling_type" in params:
            request["type_filling"] = filling_type[params["filling_type"]]

        if "time_type" in params:
            request["type_time"] = filling_type[params["time_type"]]

        if "price" in params:
            request["price"] = params["price"]

        if "lots" in params:
            request["volume"] = float(params["lots"])

        if "stoploss" in params:
            request["sl"] = params["stoploss"]

        if "takeprofit" in params:
            request["tp"] = params["takeprofit"]

        if "ticket" in params:
            request["order"] = params["ticket"]

        if "position" in params:
            request["position"] = params["position"]

        if "expire_time" in params:
            request["expiration"] = params["expire_time"]

        if "deviation" in params:
            request["deviation"] = params["deviation"]

        result = mt5.order_send(request)

        if result is None:

            if self.debug:
                print("[Log`]: Unable to add order, check provided parameters")
                print("Request:", pd.Series(request))

            return {
                'magic_id': self.magic,
                'identifier': id,
                'result': None,
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }

        result = result._asdict()

        if result["retcode"] == 10019:

            if self.debug:
                logging.warn(f"Code [{result['retcode']}]: {result['comment']}")
                logging.warn(
                    "Unable to place order. Close some positions or add some funds"
                )

            return {
                'magic_id': self.magic,
                'identifier': id,
                'result': result,
                'error_info': '[10019]: Close some positions or add funds.',
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }

        if result["retcode"] != 10009:

            if self.debug:
                logging.info(
                    f"Code [{result['retcode']}]: Error while placing Order. Please try again."
                )
                logging.debug(f"\n Request \n:  {pd.Series(request)}")
                logging.debug(f"\n Response \n:  {pd.Series(result)}")

            return {
                'magic_id': self.magic,
                'identifier': id,
                'result': result,
                'error_info': 'Check last error or result.',
                'last_error': mt5.last_error(),
                'params': params,
                '_request': request
            }

        return {
            "magic_id": self.magic,
            "identifier": id,
            "result": result,
            'error_info': None,
            'last_error': None,
            'params': params,
            '_request': request
        }
